!-------------------------------------------------------------------------------
!
!  Quasicontinuum (QC) Method: Mixed Continuum and Atomistic Simulation Package
!  QC Package distribution version 1.4 (November 2011)
!
!  Copyright (C) 2003 R. Miller, M. Ortiz, R. Phillips, D. Rodney, E. B. Tadmor
!  Copyright (C) 2004, 2005, 2006, 2007, 2011 R. Miller, E. B. Tadmor
!
!  This program is free software; you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation; either version 2 of the License, or
!  (at your option) any later version.
!
!  This program is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with this program; if not, write to the Free Software
!  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
!
!  For more information please visit the QC website at:
!
!    www.qcmethod.com
!
!  or contact Ron Miller or Ellad Tadmor with contact information below:
!
!  NOTE HOWEVER THAT NEITHER RON MILLER NOR ELLAD TADMOR WILL BE ABLE TO
!  PROVIDE SUPPORT OR DEBUGGING SERVICES FOR THE CODE. THE CODE IS SUPPLIED 
!  AS IS FOR YOUR USE.
!
!    Ron Miller
!    Department of Mechanical and Aerospace Engineering
!    Carleton University
!    1125 Colonel By Drive
!    Ottawa, ON, K1S 5B6
!    CANADA
!   
!    URL: http://www.mae.carleton.ca/Ron_Miller
!
!  -or-
!
!    Ellad B. Tadmor
!    Department of Aerospace Engineering and Mechanics
!    University of Minnesota
!    107 Akerman Hall, 110 Union St SE
!    Minneapolis, MN 55455
!    USA
!    
!    URL: http://www.aem.umn.edu/people/faculty/bio/Tadmor.shtml
!   
!-------------------------------------------------------------------------------
!****************************************************************************
!**
!**  MODULE mod_uservars   contains variables used by user routines
!**
!**                        Grain boundary example
!**
!**  Variable Definitions:
!**  ---------------------
!**
!**  Integer Variables:
!**
!**    user_iparam          -  array of integers available for storage
!**                            of user-specified parameters. See also
!**                            user_rparam.
!**
!**  Real Variables:
!**
!**    user_rparam          -  array of reals available for storage
!**                            of user-specified parameters. See also
!**                            user_iparam.
!**
!****************************************************************************

      module mod_uservars
      use mod_global, only : dp

      save
      private
      public user_iparam, user_rparam

      integer user_iparam(10)
      real(kind=dp) user_rparam(10)

      end module mod_uservars


!**-----------------------------------------------------------------------------
!**
!**   Mesh generator routine for GB example
!**   Creates an initial uniform mesh.
!**
!**   Calling format:
!**
!**   mesh,,nx,ny
!**
!**   where  nx = number of x-divisions
!**          ny = number of y-divisions
!--
      subroutine user_mesh(id,x,ix,f,b,itx,key,input)
      use mod_uservars
      use mod_global,   only : dp,maxel,maxnp,ndf,ndm,nen,numel,numnp,
     &                         nxdm,SymmetricMesh
      use mod_output,   only : erroroutput,nerror,output,string,error
      use mod_boundary, only : ncb,nce,NCEMAX,elist
      use mod_repatom,  only : protected
      use mod_qclib,    only : next,freein,qsortr
      use mod_mesh,     only : delaunay,PerturbMesh
      use mod_grain,    only : NumGrains,GetGrainNumVrts,GetGrainVertex,
     &                         NearestBSite
      implicit none

!-- Transferred variables
      integer,           intent(inout)  :: id(ndf,maxnp),
     &                                     ix(nen,maxel),
     &                                     itx(3,maxel)
      real(kind=dp),     intent(inout)  :: x(nxdm,maxnp),
     &                                     f(ndf,maxnp),
     &                                     b(ndf,maxnp)
      character(len=4),  intent(in)     :: key
      character(len=80), intent(in)     :: input

!-- Local variables
      real(kind=dp), allocatable :: vtmp(:)
      integer nx,ny,kount,igr,i,j,idum,ivert,lower,upper,numvrts
      real(kind=dp) xlo,xhi,ylo,yhi,dx,dy,xx,yy,xtmp(3),xtmp1(2),
     & xmin,xmax,ymin,ymax,xc,yc,dum,vertex(ndm)
      real(kind=dp), parameter :: tol=1.e-3_dp

      ! Parse input
      lower = 0
      upper = next(lower,input)
      call freein(input,lower,upper,nx,dum,1)
      lower = upper
      upper = next(lower,input)
      call freein(input,lower,upper,ny,dum,1)


      string=''; call output
      write(string,'(a)')
     &     'Generating uniform mesh for grain boundary test problem'
      call output
      string=''; call output
      write(string,'(a,i4)')   '# x-elements           = ',nx
      call output
      write(string,'(a,i4)')   '# y-elements           = ',ny
      call output

      xlo=1.e30_dp
      xhi=-1.e30_dp
      ylo=1.e30_dp
      yhi=-1.e30_dp
      do igr=1,NumGrains()
         call GetGrainNumVrts(igr,numvrts)
         do ivert=1,numvrts
            call GetGrainVertex(igr,ivert,vertex)
            xx = vertex(1)
            yy = vertex(2)
            xlo = min(xx,xlo)
            xhi = max(xx,xhi)
            ylo = min(yy,ylo)
            yhi = max(yy,yhi)
         enddo
      enddo
      dx = (xhi - xlo)/real(nx,dp)
      dy = (yhi - ylo)/real(ny,dp)

      ! Generate Nodes
      kount = 0
      do i = 0,nx
      do j = 0,ny
         xtmp1(1) = xlo + real(i,dp)*dx
         xtmp1(2) = ylo + real(j,dp)*dy
         call NearestBSite(xtmp1,.false.,xtmp,igr,idum,x,b,ix,itx)
         kount = kount + 1
         if (kount.gt.maxnp) then
            nerror = 1
            error(1) =
     &         'Number of nodes exceeds maxnp in subroutine user_mesh'
            call erroroutput
            stop
         endif
         x(1:3,kount) = xtmp(1:3)
      enddo
      enddo
      numnp = kount

      ! Identify model boundaries
      xmin = x(1,1)
      xmax = x(1,1)
      ymin = x(2,1)
      ymax = x(2,1)
      do i = 1,numnp
         xx = x(1,i)
         yy = x(2,i)
         if (xx.lt.xmin) xmin=xx
         if (xx.gt.xmax) xmax=xx
         if (yy.lt.ymin) ymin=yy
         if (yy.gt.ymax) ymax=yy
      enddo
      xc = (xmin+xmax)/2.0_dp
      yc = (ymin+ymax)/2.0_dp

      ! Remove duplicate nodes
      i = 1
      do while (i<numnp)
         j = i + 1
         do while (j<=numnp)
            if (abs( x(1,i) - x(1,j)) .lt. tol .and.
     &          abs( x(2,i) - x(2,j)) .lt. tol ) then
               x(:,j)=x(:,numnp)
               numnp = numnp - 1
            else
               j = j + 1
            endif
         enddo
         i = i + 1
      enddo

      ! Identify nodes lying on the boundaries and sort 
      ! in counterclockwise fashion for constrained Delaunay algorithm
      xmax=xmax-0.2_dp*dx
      xmin=xmin+0.2_dp*dx
      ymax=ymax-0.2_dp*dy
      ymin=ymin+0.2_dp*dy
      nce = 0
      allocate(vtmp(numnp))
      do i=1,numnp
         xx=x(1,i)
         yy=x(2,i)
         if (xx.lt.xmin .or. xx.gt.xmax .or.
     &       yy.lt.ymin .or. yy.gt.ymax) then
            nce=nce+1
            if (nce.gt.NCEMAX) then
               nerror = 1
               error(1) = 'nce exceeds NCEMAX in subroutine user_mesh'
               call erroroutput
               stop
            endif
            elist(1,nce)=i
            vtmp(i)=datan2(yy-yc,xx-xc)
         endif
      enddo
      call qsortr(nce,elist,vtmp,1.0_dp,NCEMAX,numnp)
      deallocate(vtmp)    

      ! Finish defining boundary
      elist(2,nce)=elist(1,1)
      ncb = nce
      do i=1,nce-1
         elist(2,i)=elist(1,i+1)
      enddo

      ! Triangulate
      SymmetricMesh = .false.   ! turn off symmetric meshing option
      call PerturbMesh(x,.true.)
      call delaunay(x,ix,b,f,id,itx)
      call PerturbMesh(x,.false.)

      ! Boundary conditions
      do i = 1,numnp
         xx=x(1,i)
         yy=x(2,i)
         ! Protect initial nodes from deletion during coarsening
         protected(i) = .true.
         ! Rigid boundary at y=ymax
         if (yy.gt.ymax) then
            id(1,i)=1
            id(2,i)=1
            id(3,i)=1
         endif
         ! Rigid boundary at y=ymin
         if (yy.lt.ymin) then
            id(1,i)=1
            id(2,i)=1
            id(3,i)=1
         endif
         ! Rigid boundary at x=xmin
         if (xx.lt.xmin) then
            id(1,i)=1
            id(2,i)=1
            id(3,i)=1
         endif
         ! Rigid boundary at x=xmax
         if (xx.gt.xmax) then
            id(1,i)=1
            id(2,i)=1
            id(3,i)=1
         endif
      enddo

      ! Report and exit
      string=''; call output
      write(string,'(a,i5)')'**numnp = ',numnp
      call output
      write(string,'(a,i5)')'**numel = ',numel
      call output

      return
      end

!**---------------------------------------------------------------
!**   user_bcon : incremental uniform shear in x-direction
!**
!--
      subroutine user_bcon(id,x,ix,f,jdiag,str,eps,q,b,dr,db,shp,
     &   xsj,key,input,flag)
      use mod_uservars
      use mod_global, only : dp,maxel,maxneq,maxnp,ndf,nen,nq,nstr,
     &                       numnp,nxdm
      use mod_output, only : output,string
      use mod_pload,  only : GetLoadPropFact,GetLoadOldPropFact,
     &                       GetLoadTime,GetLoadTimeStep
      implicit  none

!-- Transferred variables
      real(kind=dp),     intent(inout) :: b(ndf,maxnp),str(nstr,maxel),
     &                                    eps(nstr,maxel),q(nq,maxel),
     &                                    x(nxdm,maxnp),f(ndf,maxnp),
     &                                    dr(ndf,maxnp),db(maxneq),
     &                                    shp(3,nen,maxel),xsj(maxel)
      integer,           intent(inout) :: jdiag(maxneq),ix(nen,maxel),
     &                                    id(ndf,maxnp)
      character(len=80), intent(in)    :: input
      character(len=4),  intent(in)    :: key
      logical,           intent(inout) :: flag

!-- Local variables
      integer i,j
      real(kind=dp) propfact,propolfact

      propfact = GetLoadPropFact()
      propolfact = GetLoadOldPropFact()
      do i=1,numnp
         b(1,i)=b(1,i)+(propfact-propolfact)*x(2,i)
         if(id(1,i)==1) f(1,i)=x(2,i)
      enddo

      ! output
      write(string,'(a)')'* Boundary conditions applied:'
      call output
      write(string,'(2x,a,e15.6)')
     &             'uniform shear in x-direction ',propfact
      call output
      return
      end

!**---------------------------------------------------------------
!**   user_pdel : compute the total external load for the p-delta 
!**               curve and for convergence checks.
!**
!--
      subroutine user_pdel(id,x,ix,f,jdiag,str,eps,q,b,dr,db,shp,
     &   xsj,key,input,force)
      use mod_uservars
      use mod_global, only : dp,maxel,maxneq,maxnp,ndf,nen,nq,nstr,
     &                       numnp,nxdm
      implicit  none

!-- Transferred Variables
      real(kind=dp),     intent(inout) :: b(ndf,maxnp),str(nstr,maxel),
     &                                    eps(nstr,maxel),q(nq,maxel),
     &                                    x(nxdm,maxnp),f(ndf,maxnp),
     &                                    dr(ndf,maxnp),db(maxneq),
     &                                    shp(3,nen,maxel),xsj(maxel)
      integer,           intent(inout) :: jdiag(maxneq),ix(nen,maxel),
     &                                    id(ndf,maxnp)
      character(len=80), intent(in)    :: input
      character(len=4),  intent(in)    :: key
      real(kind=dp),     intent(out)   :: force

!-- Local Variables
      integer i
 
      force = 0.0_dp
      do i = 1,numnp
         if (id(1,i)==1 .and. x(2,i)>0.0_dp) force=force-dr(1,i) 
      enddo

      return
      end

!**---------------------------------------------------------------
!**   user_potential : user-defined external potential 
!**
!**
!**
      subroutine user_potential(id,x,ix,f,jdiag,str,eps,q,b,dr,shp,xsj,
     &                          fls,flt,flw)
      use mod_uservars
      use mod_global, only : dp,maxneq,maxnp,maxel,ndf,nen,nq,nstr,nxdm,
     &                       user_energy
      use mod_pload,  only : GetLoadPropFact,GetLoadOldPropFact,
     &                       GetLoadTime,GetLoadTimeStep
      use mod_stiff,  only : tang,istiff
      implicit  none

!-- Transferred Variables
      real(kind=dp), intent(inout) :: b(ndf,maxnp),str(nstr,maxel),
     &                                eps(nstr,maxel),q(nq,maxel),
     &                                x(nxdm,maxnp),f(ndf,maxnp),
     &                                dr(ndf,maxnp),shp(3,nen,maxel),
     &                                xsj(maxel)
      integer,       intent(inout) :: jdiag(maxneq),ix(nen,maxel),
     &                                id(ndf,maxnp)
      logical,       intent(in)    :: fls,flt,flw

!-- Local Variables

      user_energy = 0.0_dp

      return
      end

!--------------------------------------------------------------------
!
!     user_plot : User-specified output plot
!
!     Calling format:
!
!     plot,user,filepref,[index],[umag],[scale],[append]
!
!
!
      subroutine user_plot(id,x,ix,itx,f,jdiag,str,eps,q,b,dr,db,shp,
     &                     xsj,index,umag,scale,logic,input,upper)
      use mod_uservars
      use mod_global, only : dp,maxnp,maxel,maxneq,ndf,nen,nq,nstr,nxdm
      implicit  none

!-- Transferred Variables
      real(kind=dp),     intent(in)    :: b(ndf,maxnp),str(nstr,maxel),
     &                                    eps(nstr,maxel),q(nq,maxel),
     &                                    x(nxdm,maxnp),f(ndf,maxnp),
     &                                    dr(ndf,maxnp),db(maxneq),
     &                                    shp(3,nen,maxel),xsj(maxel),
     &                                    umag,scale
      integer,           intent(in)    :: jdiag(maxneq),ix(nen,maxel),
     &                                    id(ndf,maxnp),itx(3,maxel),
     &                                    index,logic
      integer,           intent(inout) :: upper
      character(len=80), intent(in)    :: input

      return
      end
